.packageName <- "mtm"

.First.lib <- function(lib, pkg) library.dynam("MTM_COMBO", pkg, lib)

mtm.lag1<-function(marginal, trans1, id, beta=NULL, alpha1=NULL, offset=NULL, data = sys.frame(sys.parent()), tol = 1e-4)
{
#
# Maximum likelihood estimation for lag-1 marginalized transition models.
# (basic specification given by Azzalini)
#
	
  mf <- model.frame(marginal, data)
  terms <- attr(mf, "terms")
  tl.x <- attr(terms, "term.labels")
  intr <- attr(terms, "intercept")
  if(intr > 0) tl.x <- c("(Intercept)", tl.x)
  y <- model.response(mf, "numeric")
  n <- length(y)
  x <- model.matrix(marginal, mf, contrasts)
  
  mf <- model.frame(trans1, data)
  terms <- attr(mf, "terms")
  tl.z <- attr(terms, "term.labels")
  intr.z <- attr(terms, "intercept")
  if(intr.z > 0) tl.z <- c("(Intercept)", tl.z)
  z <- model.matrix(trans1, mf)

  if(is.null(beta)) {
    # Provide a reasonable estimate for beta:
	fitbeta <- glm(y ~ -1 + x, family=binomial)
	beta <- fitbeta$coef
	
  } else {
  
    if(length(beta) != length(x[1,])) {stop("Number of estimates in beta is ", length(beta), 
			" should equal ", length(x[1,]), " (number of marginal covariates).")
	}
  }
  p <- length(x[1,])
  bnames <- tl.x
  if( is.null( bnames ) ) bnames <- paste("beta",1:p)

  anames <- tl.z
  q <- length(z[1,])
  if( is.null( anames ) ) anames <- paste("alpha1",1:q)
 
  if(is.null(alpha1)) {
  
	# Provide a reasonable estimate for alpha1:
	w1 <- as.vector(unlist(by(as.vector(unlist(by(c(1:n), id, cumsum))), id, diff)))
	w0 <- w1-1
	alpha1 <- coef(glm(y[w1] ~ x[w0,] + y[w0]*z[w0,], family=binomial))
	a.n <- length(alpha1)
	if(intr.z>0) {
		alpha1 <- as.vector(c(alpha1[match("y[w0]", names(alpha1))], alpha1[(a.n - q+1):a.n]))
		alpha1 <- alpha1[!is.na(alpha1)]
	} else {
		alpha1 <- as.vector(alpha1[(a.n - q+1):a.n])
		alpha1 <- alpha1[!is.na(alpha1)]
	}
  } else {
    if(length(alpha1) != length(z[1,])) {stop("Number of estimates in alpha1 is ", length(alpha1), 
			" should equal ", length(z[1,]), " (number of transition covariates).")
    }
  }
  	
	nclust <- length( table( id ) )
	covM<-matrix( 0, p+q, p+q )
	covE<-matrix( 0, p+q, p+q )
        logL <- 0.0
	ScoreTest <- 0.0
        muTrans <- rep( 0, n )
        if( is.null(offset) ) offset <- rep( 0.0, n )
#
	z <- .C("mtm_lag1",
		as.double(x),
		as.double(y),
		as.double(z),
		as.double(id),
		beta = as.double(beta),
                alpha = as.double(alpha1),
                logL = as.double(logL),
		as.integer(n),
		as.integer(nclust),
		as.integer(p),
		as.integer(q),
		tol=as.double(tol),
		covM=as.double(covM),
		covE=as.double(covE),
		ScoreTest=as.double(ScoreTest),
		muTrans=as.double(muTrans),
                offset=as.double(offset) )
#
        covM <- matrix(z[["covM"]],p+q,p+q)
        covE <- matrix(z[["covE"]],p+q,p+q)
#

out <- list( beta = z[["beta"]], alpha = z[["alpha"]], logL = z[["logL"]], 
		covM=covM, covE=covE, ScoreTest=z[["ScoreTest"]], 
                muTrans=z[["muTrans"]] )

names(out$beta) <- bnames
names(out$alpha) <- anames
class(out) <- "mtm1"
#
out
}
##########
print.mtm1 <- function( out ){
#
cat("\n")
cat("Marginalized Transition Models - lag 1 \n\n")
p <- length( out$beta )
q <- length( out$alpha )

#
seM.beta <- sqrt( diag( out$covM[1:p,1:p] ) )
seE.beta <- sqrt( diag( out$covE[1:p,1:p] ) )
bmat <- cbind( out$beta, seM.beta, seE.beta, out$beta/seM.beta )
dimnames( bmat ) <- list( names(out$beta), 
                          c("est.","s.e. M","s.e. E", "Z") )
#
if( q>1 ){
  seM.alpha <- sqrt( diag( out$covM[(p+1):(p+q),(p+1):(p+q)] ) )
  seE.alpha <- sqrt( diag( out$covE[(p+1):(p+q),(p+1):(p+q)] ) )
  amat <- cbind( out$alpha, seM.alpha, seE.alpha, out$alpha/seM.alpha )
  dimnames( amat ) <- list( names(out$alpha), 
                            c("est.","s.e. M","s.e. E", "Z") )
}else{
  seM.alpha <- sqrt( out$covM[p+1,p+1] )
  seE.alpha <- sqrt( out$covE[p+1,p+1] )
  amat <- matrix(cbind(out$alpha, seM.alpha, seE.alpha, out$alpha/seM.alpha), nrow=1)
  dimnames( amat ) <-  list(names(out$alpha), c("est.","s.e. M","s.e. E", "Z")) 
}

#
cat( paste("  maximized logLikelihood =", round( out$logL, 3 ), "\n\n") )
#
cat( paste("  Beta estimates:\n\n") )
print( round( bmat, 3 ) )
#
cat("\n")
cat( paste("  Alpha1 estimates:\n\n") )

print( round(amat,3) )

cat("\n")
cat( paste("  Score test for lag-2 coefficient =", round(out$ScoreTest,3),
  "p-val =", round( 1-pchisq( out$ScoreTest, df=1 ), 3 ),"\n") )
cat("\n")
}

mtm.lag2<-function( marginal, trans1, trans2, id, beta=NULL, alpha1=NULL, alpha2=NULL, 
                    tol = 1e-4, iter = 50, data=sys.frame(sys.parent()))
{
#
# Maximum likelihood estimation for lag-2 marginalized transition models.
# 
  mf <- model.frame(marginal, data)
  terms <- attr(mf, "terms")
  tl.x <- attr(terms, "term.labels")
  intr <- attr(terms, "intercept")
  if(intr > 0) tl.x <- c("(Intercept)", tl.x)
  y <- model.response(mf, "numeric")
  n <- length(y)
  x <- model.matrix(marginal, mf, contrasts)
  
  mf <- model.frame(trans1, data)
  terms <- attr(mf, "terms")
  tl.z1 <- attr(terms, "term.labels")
  intr.z1 <- attr(terms, "intercept")
  if(intr.z1 > 0) tl.z1 <- c("(Intercept)", tl.z1)
  z1 <- model.matrix(trans1, mf)

  mf <- model.frame(trans2, data)
  terms <- attr(mf, "terms")
  tl.z2 <- attr(terms, "term.labels")
  intr.z2 <- attr(terms, "intercept")
  if(intr.z2 > 0) tl.z2 <- c("(Intercept)", tl.z2)
  z2 <- model.matrix(trans2, mf)	
	
  if(is.null(beta)) {
    # Provide a reasonable estimate for beta:
	fitbeta <- glm(y ~ -1 + x, family=binomial)
	beta <- fitbeta$coef
  } else {
  
    if(length(beta) != length(x[1,])) {stop("Number of estimates in beta is ", length(beta), 
			" should equal ", length(x[1,]), " (number of marginal covariates).")
	}
  }
  p <- length(x[1,])
  bnames <- tl.x
  if( is.null( bnames ) ) bnames <- paste("beta:",1:p)

## If alpha1 is NULL, then generate some estimates:
  a1names <- tl.z1
  q1 <- length(z1[1,])
  if( is.null( a1names ) ) a1names <- paste("alpha1:",1:q1)
  if(is.null(alpha1)) {
	# Provide a reasonable estimate for alpha1:
	w1 <- as.vector(unlist(by(as.vector(unlist(by(c(1:n), id, cumsum))), id, diff)))
	w0 <- w1-1
	alpha1 <- coef(glm(y[w1] ~ x[w0,] + y[w0]*z1[w0,], family=binomial))
	a1.n <- length(alpha1)
	if(intr.z1>0) {
		alpha1 <- as.vector(c(alpha1[match("y[w0]", names(alpha1))], alpha1[(a1.n - q1+1):a1.n]))
		alpha1 <- alpha1[!is.na(alpha1)]
	} else {
		alpha1 <- as.vector(alpha1[(a1.n - q1+1):a1.n])
		alpha1 <- alpha1[!is.na(alpha1)]
	}
  } else {
    if(length(alpha1) != length(z1[1,])) {stop("Number of estimates in alpha1 is ", length(alpha1), 
			" should equal ", length(z1[1,]), " (number of transition-1 covariates).")
    }
  }
  
## If alpha2 is NULL, then generate some estimates:
  a2names <- tl.z2
  q2 <- length(z2[1,])
  if( is.null( a2names ) ) a2names <- paste("alpha2:",1:q2)
  if(is.null(alpha2)) {
	ret.w2 <- function(wvec) { if(length(wvec)<3) { return(NULL)
								}else {diff(cumsum(wvec))[2:(length(wvec)-1)] }
	}
	w2 <- as.vector(unlist(by(c(1:n), id, ret.w2)))
	w0 <- w2 - 2
	alpha2 <- coef(glm(y[w2] ~ x[w0,] + y[w0]*z2[w0,], family=binomial))
	a2.n <- length(alpha2)
	if(intr.z2>0) {
		alpha2 <- as.vector(c(alpha2[match("y[w0]", names(alpha2))], alpha2[(a2.n - q2+1):a2.n]))
		alpha2 <- alpha2[!is.na(alpha2)]
	} else {
		alpha2 <- as.vector(alpha2[(a2.n - q2+1):a2.n])
		alpha2 <- alpha2[!is.na(alpha2)]
	}
  } else {
    if(length(alpha2) != length(z2[1,])) {stop("Number of estimates in alpha2 is ", length(alpha2), 
			" should equal ", length(z2[1,]), " (number of transition-2 covariates).")
    }
  }


  nclust <- length( table( id ) )
  covM<-matrix( 0, p+q1+q2, p+q1+q2 ) 
  covE<-matrix( 0, p+q1+q2, p+q1+q2 )
  logL <- 0.0
  ScoreTest <- 0.0
  muTrans <- rep( 0, n )
  TheIntegers <- c( p, q1, q2, n, nclust )
  
  z <- .C("mtm_lag2",
		as.double(x),
		as.double(y),
		as.double(z1),
		as.double(z2),
		as.double(id),
		beta = as.double(beta),
		alpha1 = as.double(alpha1),
		alpha2 = as.double(alpha2),
		logL = as.double(logL),
		as.integer(TheIntegers),
		tol=as.double(tol),
		covM=as.double(covM),
		covE=as.double(covE),
		ScoreTest=as.double(ScoreTest),
		muTrans=as.double(muTrans),
		as.double(iter))
#
  covM <- matrix(z[["covM"]],p+q1+q2,p+q1+q2)
  covE <- matrix(z[["covE"]],p+q1+q2,p+q1+q2)
#
  out <- list( beta = z[["beta"]], alpha1 = z[["alpha1"]], 
                alpha2=z[["alpha2"]], logL = z[["logL"]], 
		covM=covM, covE=covE, ScoreTest=z[["ScoreTest"]], 
                muTrans=z[["muTrans"]] )

  names(out$beta) <- bnames
  names(out$alpha1) <- a1names
  names(out$alpha2) <- a2names
class(out) <- "mtm2"
#
out
}
##########

print.mtm2 <- function( out ){
#
cat("\n")
cat("Marginalized Transition Models - lag 2 \n\n")
p <- length( out$beta )
q1 <- length( out$alpha1 )
q2 <- length( out$alpha2 )
#
seM.beta <- sqrt( diag( out$covM[1:p,1:p] ) )
seE.beta <- sqrt( diag( out$covE[1:p,1:p] ) )
bmat <- cbind( out$beta, seM.beta, seE.beta, out$beta/seM.beta )
dimnames( bmat ) <- list( names(out$beta), 
                          c("est.","s.e. M","s.e. E", "Z") )
#
if( q1>1 ){
  seM.alpha <- sqrt( diag( out$covM[(p+1):(p+q1),(p+1):(p+q1)] ) )
  seE.alpha <- sqrt( diag( out$covE[(p+1):(p+q1),(p+1):(p+q1)] ) )
  amat <- cbind( out$alpha1, seM.alpha, seE.alpha, out$alpha1/seM.alpha )
  dimnames( amat ) <- list( names(out$alpha1), 
                            c("est.","s.e. M","s.e. E", "Z") )
}else{
  seM.alpha <- sqrt( out$covM[p+1,p+1] )
  seE.alpha <- sqrt( out$covE[p+1,p+1] )
  amat <- matrix(cbind( out$alpha1, seM.alpha, seE.alpha, out$alpha1/seM.alpha ), nrow=1)
  dimnames( amat ) <-  list(names(out$alpha1), c("est.","s.e. M","s.e. E", "Z")) 
}
a1mat <- amat
#
if( q2>1 ){
  seM.alpha <- sqrt( diag( out$covM[(p+q1+1):(p+q1+q2),(p+q1+1):(p+q1+q2)] ) )
  seE.alpha <- sqrt( diag( out$covE[(p+q1+1):(p+q1+q2),(p+q1+1):(p+q1+q2)] ) )
  amat <- cbind( out$alpha2, seM.alpha, seE.alpha, out$alpha2/seM.alpha )
  dimnames( amat ) <- list( names(out$alpha2), 
                            c("est.","s.e. M","s.e. E", "Z") )
}else{
  seM.alpha <- sqrt( out$covM[p+q1+1,p+q1+1] )
  seE.alpha <- sqrt( out$covE[p+q1+1,p+q1+1] )
  amat <- matrix(cbind(out$alpha2, seM.alpha, seE.alpha, out$alpha2/seM.alpha ), nrow=1)
  dimnames( amat ) <-  list(names(out$alpha2), c("est.","s.e. M","s.e. E", "Z"))
}
a2mat <- amat
#
cat( paste("  maximized logLikelihood =", round( out$logL, 3 ), "\n\n") )
#
cat( paste("Beta estimates:\n\n") )
print( round( bmat, 3 ) )
#
cat("\n")
cat( paste("Alpha1 estimates:\n\n") )
print( round( a1mat, 3 ) )
cat("\n")
cat( paste("Alpha2 estimates:\n\n") )
print( round( a2mat, 3 ) )
cat("\n")
cat( paste("Score test for lag-3 coefficient =", round(out$ScoreTest,3),
  "p-val =", round( 1-pchisq( out$ScoreTest, df=1 ), 3 ),"\n") )
cat("\n")
}


