/*  lag-1 and lag-2 Marginalized Transition Models					*/
/*                                                                  */
/*                              Created by:		P. Heagerty         */
/*                                              00/06/21            */
/*																	*/
/*								Updated by:		B. Comstock			*/
/*												2006/08/27			*/
/*                                                                  */

#include "chanmatstruct.h"
#include "chanmatfuns.h"

#define cfree free

#define HALFSTEP2 0

#define EPSILON2 1e-8
#define MAX_COUNT 50
#define NearZero 1e-8
#define RIDGE 1e-4

#define MAXITS 100		/* From lag-1 */
#define LAMBDA 1.0
#define HALFSTEP1 1
#define EPSILON1 1e-5

extern double  matmaxabs(), antilogit(), logit(), BoundaryCheck();

extern MATRIX  **get_score_calcs1();
extern MATRIX  **get_score_calcs2();

void mtm_lag2( S_x, S_y, S_z1, S_z2, S_id, S_beta, S_alpha1, S_alpha2, S_logL,
	     S_Integers, S_tol, S_covM, S_covE, S_test, S_muT, S_iter )

  double  *S_x, *S_y, *S_z1, *S_z2, *S_id, *S_beta;
  double  *S_alpha1, *S_alpha2, *S_covM, *S_covE;
  double  *S_test, *S_muT, *S_logL, *S_tol, *S_iter;
  integer *S_Integers;
{
  MATRIX  **Y, **X, **Z1, **Z2, **MuTrans;
  MATRIX  *y, *mu, *beta, *alpha1, *alpha2, *z1, *z2, *x, *id, *muT;
  double  dmax, tolerance, logL, logLc, ccc, ScoreTest, U2, V2, U2c, V2c;
  MATRIX  *delta, *DVD, *DVR;
  MATRIX  **M ;
  MATRIX  *UUt, *Ie, *Im;
  int     i, j, k, n, nobs, nclust, p, q1, q2, one, nc2, count, iter, c;
  int     method, converge, error_flag;
  FILE    *wfile;


  one = 1;

  p = *(S_Integers+0);
  q1 = *(S_Integers+1);
  q2 = *(S_Integers+2);
  nobs = *(S_Integers+3);
  nclust = *(S_Integers+4);
  tolerance = *S_tol;

  logL = *S_logL;

  wfile = fopen("MLE.out","w");
  fprintf( wfile, "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
  fprintf( wfile, " p = %i, q1 = %i, q2 = %i, n = %i, tol = %f\n", 
                    p, q1, q2, nobs, tolerance );
  fprintf( wfile, "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n\n" );
  fclose( wfile );
  
  from_S( S_beta, &p, &one, beta );
  make_permanent( beta );

  from_S( S_alpha1, &q1, &one, alpha1 );
  make_permanent( alpha1 );

  from_S( S_alpha2, &q2, &one, alpha2 );
  make_permanent( alpha2 );

  from_S( S_x, &nobs, &p, x );
  from_S( S_y, &nobs, &one, y );
  from_S( S_z1, &nobs, &q1, z1 );
  from_S( S_z2, &nobs, &q2, z2 );
  from_S( S_muT, &nobs, &one, muT);
  from_S( S_id, &nobs, &one, id );

  X = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  Y = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  Z1 = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  Z2 = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  MuTrans = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  
  make_permanent( id );

  split( x, id, X );
  split( y, id, Y );
  split( z1, id, Z1 );
  split( z2, id, Z2 );
  split( muT, id, MuTrans );

  destroy_matrix( x );
  destroy_matrix( y );
  destroy_matrix( z1 );
  destroy_matrix( z2 );
  destroy_matrix( id );
  destroy_matrix( muT );

  for(c=0;c<nclust;c++){
    make_permanent( Y[c] );
    make_permanent( X[c] );
    make_permanent( Z1[c] );
    make_permanent( Z2[c] ); }


  iter=0;
  converge=0;


  do{

    DVR = create_matrix( p+q1+q2, 1, EPHEMERAL );
    DVD = create_matrix( p+q1+q2, p+q1+q2, EPHEMERAL );
    UUt = create_matrix( p+q1+q2, p+q1+q2, EPHEMERAL );

    logL = 0.0;
    U2 = 0.0;
    V2 = 0.0;

    for(c=0;c<nclust;c++){

      n = Y[c]->nrows;

      logLc = 0.0;
      U2c = 0.0;
      V2c = 0.0;

      error_flag = 0;

      M = get_score_calcs2( Y[c], X[c], Z1[c], Z2[c], beta, alpha1, alpha2,
                           n, &logLc, &U2c, &V2c, MuTrans[c], &error_flag  );

      if( error_flag>0 ){
	printf(" Errors for cluster = %i, error_flag = %i \n", c, error_flag);
      }

      make_permanent( M[0] );
      UUt = matadd( UUt, matmult( M[0], transp( M[0] ) ) );
      make_ephemeral( M[0] );
      DVR = matadd( DVR, M[0] );
      DVD = matadd( DVD, M[1] );

      cfree( M );
  
      logL += logLc;
      U2 += U2c;
      /*      V2 += V2c; */
      V2 += U2c*U2c;

    }/* c */

    for( j=0; j<(p+q1+q2); j++ ) MEL(DVD,j,j) += RIDGE;

    make_permanent( DVD );
    make_permanent( DVR );

    delta = matmult( luinv( DVD ), DVR );

wfile=fopen("MLE.out","a");
fprintf(wfile," score\n");
fmatdump(wfile, DVR );
fprintf(wfile," hessian\n");
fmatdump(wfile,DVD);
fprintf(wfile," delta\n");
fmatdump(wfile,delta);
fprintf(wfile," logL = %f\n", logL );
fclose(wfile); 

    make_permanent( delta );
    dmax = matmaxabs( delta );

    ccc = 1.0;
    if( HALFSTEP2 && iter<10 ) ccc = 0.5;

    for( j=0; j<p; j++) MEL( beta, j, 0 ) += ccc*MEL( delta, j, 0 );
    for( j=0; j<q1; j++) MEL( alpha1, j, 0 ) += ccc*MEL( delta, p+j, 0 );
    for( j=0; j<q2; j++) MEL( alpha2, j, 0 ) += ccc*MEL( delta, p+q1+j, 0 );
    destroy_matrix( delta );

    if( dmax < tolerance ){
      converge = 1;
    }else{
      destroy_matrix( DVD );
      destroy_matrix( DVR );
      destroy_matrix( UUt );
    }

    iter++;

 }while( (iter < *S_iter) && (!converge) );                     


  /* add the appropriate cleaning... */
  count = 0;
  for( i=0; i<nclust; i++ ){
    for( j=0; j<MuTrans[i]->nrows; j++ ){
      *(S_muT+count) = MEL( MuTrans[i], j, 0 );
      count++;
    }
    destroy_matrix( MuTrans[i] );
    destroy_matrix( Y[i] );
    destroy_matrix( X[i] );
    destroy_matrix( Z1[i] );
    destroy_matrix( Z2[i] );
  }
  cfree( MuTrans );
  cfree( Y );
  cfree( X );
  cfree( Z1 );
  cfree( Z2 );

  to_S( beta, S_beta );

  to_S( alpha1, S_alpha1 );

  to_S( alpha2, S_alpha2 );

  if( converge ){

    Im = luinv( DVD );
    make_permanent( Im );

    Ie = matmult( Im, matmult( UUt, transp(Im) ) );

    to_S( Im, S_covM );

    to_S( Ie, S_covE );

  }

  *S_logL = logL;

  ScoreTest = U2*U2/V2;

  *S_test = ScoreTest;

}/* end of routine */

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

MATRIX   **get_score_calcs2( y, x, z1, z2, beta, alpha1, alpha2,
                            n, logLc, U2c, V2c, muTc, error_flag )

  MATRIX *y, *x, *z1, *z2, *beta, *alpha1, *alpha2, *muTc;
  double *logLc, *U2c, *V2c;
  int    n, *error_flag;

{
  MATRIX *score, *hessian;
  MATRIX **M, *mu, *psi1, *psi2, *lpM, *lp1, *lp2;
  MATRIX *dp00dbeta, *dp00dalpha1, *dp00dalpha2, *dnew00;
  MATRIX *dp01dbeta, *dp01dalpha1, *dp01dalpha2, *dnew01;
  MATRIX *dp10dbeta, *dp10dalpha1, *dp10dalpha2, *dnew10;
  MATRIX *dp11dbeta, *dp11dalpha1, *dp11dalpha2, *dnew11;
  MATRIX *Bt, *Ct, *Dt, *Et;
  double  At;
  double mu0, mu1, mu2, psi01, psi02, y0, y1, y2, y3;
  double dmu0dJ, dmu1dJ, dDeltaJ, dDeltaK;
  double delta0, gamma1, gamma2;
  double p00, p01, p10, p11, lpC00, lpC01, lpC10, lpC11;
  double new00, new01, new10, new11;
  double mu_current;
  double A, B, C, R, muTrans;
  double dp11dmu1, dp11dmu0, dp11dpsi01;
  double dlogLcdmu1, dlogLcdmu0, dlogLcdpsi01;
  double dmuTransdmu1, dmuTransdmu0, dmuTransdpsi01;
  double dmu1dthetaJ, dmu0dthetaJ, dpsi01dthetaJ;
  double dmu1dthetaK, dmu0dthetaK, dpsi01dthetaK;
  double dJ, dK;
  double dstep;
  double logLikelihood;
  int    p, q1, q2, i, j, k;
  int    count, converge_delta;
  FILE   *wfile;

  p = beta->nrows;
  q1 = alpha1->nrows;
  q2 = alpha2->nrows;

  /* create the score and hessian matrices */

  score = create_matrix( p+q1+q2, 1, EPHEMERAL );
  hessian = create_matrix( p+q1+q2, p+q1+q2, EPHEMERAL );

  /* create the history derivative vectors */

  dp00dbeta = create_matrix( p, 1, EPHEMERAL );
  dp01dbeta = create_matrix( p, 1, EPHEMERAL );
  dp10dbeta = create_matrix( p, 1, EPHEMERAL );
  dp11dbeta = create_matrix( p, 1, EPHEMERAL );

  dp00dalpha1 = create_matrix( q1, 1, EPHEMERAL );
  dp01dalpha1 = create_matrix( q1, 1, EPHEMERAL );
  dp10dalpha1 = create_matrix( q1, 1, EPHEMERAL );
  dp11dalpha1 = create_matrix( q1, 1, EPHEMERAL );

  dp00dalpha2 = create_matrix( q2, 1, EPHEMERAL );
  dp01dalpha2 = create_matrix( q2, 1, EPHEMERAL );
  dp10dalpha2 = create_matrix( q2, 1, EPHEMERAL );
  dp11dalpha2 = create_matrix( q2, 1, EPHEMERAL );

  dnew00 = create_matrix( p+q1+q2, 1, EPHEMERAL );
  dnew01 = create_matrix( p+q1+q2, 1, EPHEMERAL );
  dnew10 = create_matrix( p+q1+q2, 1, EPHEMERAL );
  dnew11 = create_matrix( p+q1+q2, 1, EPHEMERAL );

  logLikelihood = 0.0;

  /* deal with scalar quantities */
  
  mu = create_matrix( n, 1, EPHEMERAL );
  psi1 = create_matrix( n, 1, EPHEMERAL );
  psi2 = create_matrix( n, 1, EPHEMERAL );

  lpM = matmult( x, beta );
  lp1 = matmult( z1, alpha1 );
  lp2 = matmult( z2, alpha2 );
  
  for(i=0;i<n;i++){
    MEL( mu, i, 0 ) = antilogit( MEL(lpM,i,0) );
    MEL( psi1, i, 0 ) = exp( MEL(lp1,i,0) );
    MEL( psi2, i, 0 ) = exp( MEL(lp2,i,0) );
  }

  destroy_matrix( lpM );
  destroy_matrix( lp1 );
  destroy_matrix( lp2 );

  /* single clusters */

  if( n==1 ){

    y0 = MEL( y, 0, 0 );
    mu0 = MEL( mu, 0, 0 );

    for( j=0; j<p; j++ ){
      MEL( score, j, 0 ) += MEL(x,0,j)*( y0 - mu0 );
      for( k=0; k<p; k++){
	MEL( hessian, j, k ) += MEL(x,0,j)*MEL(x,0,k)*mu0*(1.-mu0);
      }
    }

    logLikelihood += log(mu0)*y0 + log(1.-mu0)*(1.-y0);

    MEL( muTc, 0, 0 ) = mu0;

  }/* n==1 */

  if( n>1 ){

    /*  first two measurements */

    y1 = MEL( y, 0, 0 );
    mu1 = MEL( mu, 0, 0 );

    logLikelihood += log(mu1)*y1 + log(1.-mu1)*(1.-y1);

    for( j=0; j<p; j++ ){
      MEL( score, j, 0 ) += MEL(x,0,j)*(y1-mu1);
      for( k=0; k<p; k++ ){
	MEL( hessian, j, k ) += MEL(x,0,j)*MEL(x,0,k)*mu1*(1.-mu1);
      }
    }

    /* first transition */

    y0 = MEL( y, 1, 0 );
    mu0 = MEL( mu, 1, 0 );

    psi01  = MEL( psi1, 1, 0 );
    gamma1 = log( psi01 );

    if( fabs( gamma1 - 0.0 ) < EPSILON2 ){

	p11 = mu0*mu1;

	dp11dmu0 = mu1;

	dp11dmu1 = mu0;

	dp11dpsi01 = 0.0;

      }else{

	psi01 = exp( gamma1 );

        A = (psi01-1.);
	B = 1. + (mu1+mu0)*(psi01-1.);
	C = psi01*mu1*mu0;
        R = sqrt( B*B - 4.*A*C );

        p11 = ( B - R )/( 2.*A );
	
	dp11dmu1 = ( (psi01-1.) - (0.5/R)*( 2.*B*(psi01-1.) -
                      4.*A*psi01*mu0 ) )/( 2.*A ); 

	dp11dmu0 = ( (psi01-1.) - (0.5/R)*( 2.*B*(psi01-1.) -
                      4.*A*psi01*mu1 ) )/( 2.*A ); 

        dp11dpsi01 = ( B - R )*( -1./( 2*A*A ) ) +
                     ( ( mu1+mu0 - 
                       (0.5/R)*( 2.*B*(mu1+mu0) -
		        4.*mu1*mu0*( 2.*psi01 - 1.) ) ) )/( 2.*A ); 
      }

      muTrans = ( p11 / mu1 )*y1 + ( (mu0-p11)/(1.-mu1) )*(1.-y1);

      logLikelihood += log(muTrans)*y0 + log(1.-muTrans)*(1.-y0);

      MEL( muTc, 0, 0 ) = mu1;
      MEL( muTc, 1, 0 ) = muTrans;

      dmuTransdmu1 =( p11 * y1 )*( -1./( mu1*mu1 ) ) +
                    ( y1 / mu1 )*dp11dmu1 +
                    ( (mu0-p11)*(1.-y1) )*( 1./( (1.-mu1)*(1.-mu1) ) ) +
                    ( (1.-y1)/(1.-mu1) )*(-1.*dp11dmu1);

      dlogLcdmu1 = dmuTransdmu1 * ( y0 - muTrans )/( muTrans*(1.-muTrans) );


      dmuTransdmu0 = ( y1/mu1 )*dp11dmu0 + 
                     ( (1.-y1)/(1.-mu1) ) * ( 1. - dp11dmu0 );

      dlogLcdmu0 = dmuTransdmu0 * ( y0 - muTrans )/( muTrans*(1.-muTrans) );


      dmuTransdpsi01 = ( y1 / mu1 )*dp11dpsi01 - 
                       ( (1.-y1)/(1.-mu1) )*dp11dpsi01; 

      dlogLcdpsi01 = dmuTransdpsi01*( y0 - muTrans )/( muTrans*(1.-muTrans) );

      for( j=0; j<(p+q1); j++ ){
	
	if( j < p ){
	  dmu1dthetaJ = mu1*(1.-mu1)*MEL(x,0,j);
	  dmu0dthetaJ= mu0*(1.-mu0)*MEL(x,1,j);
	  dpsi01dthetaJ = 0.0;
	}else{
	  dmu1dthetaJ = 0.0; 
	  dmu0dthetaJ= 0.0;
	  dpsi01dthetaJ = psi01 * MEL(z1,1,j-p);
	}

        dJ = dmuTransdmu1 * dmu1dthetaJ +
             dmuTransdmu0 * dmu0dthetaJ +
             dmuTransdpsi01 * dpsi01dthetaJ;

	MEL( score, j, 0 ) += dlogLcdmu1 * dmu1dthetaJ +
                              dlogLcdmu0 * dmu0dthetaJ +
                              dlogLcdpsi01 * dpsi01dthetaJ;

	for( k=0; k<(p+q1); k++ ){

	  if( k < p ){
	    dmu1dthetaK = mu1*(1.-mu1)*MEL(x,0,k);
	    dmu0dthetaK = mu0*(1.-mu0)*MEL(x,1,k);
	    dpsi01dthetaK = 0.0;
	  }else{
	    dmu1dthetaK = 0.0; 
	    dmu0dthetaK = 0.0;
	    dpsi01dthetaK = psi01 * MEL(z1,1,k-p);
	  }

	  dK = dmuTransdmu1 * dmu1dthetaK +
	       dmuTransdmu0 * dmu0dthetaK +
               dmuTransdpsi01 * dpsi01dthetaK;

	  MEL( hessian, j, k ) += dJ * dK / ( muTrans*(1.-muTrans) );
	}
      }

      p11 = BoundaryCheck( p11 );
      p01 = BoundaryCheck( mu1 - p11 );
      p10 = BoundaryCheck( mu0 - p11 );
      p00 = BoundaryCheck( 1.0 - mu0 - mu1 + p11 );
      
      for( j=0; j<p; j++ ){
	dmu0dJ = mu0*(1.-mu0)*MEL(x,1,j);
	dmu1dJ = mu1*(1.-mu1)*MEL(x,0,j);
	MEL( dp00dbeta, j, 0 ) = 0.0 -
                                 dmu0dJ -
                                 dmu1dJ +
                                 dp11dmu0*dmu0dJ + dp11dmu1*dmu1dJ;
	MEL( dp01dbeta, j, 0 ) = dmu1dJ - 
                                 dp11dmu0*dmu0dJ -
                                 dp11dmu1*dmu1dJ;
	MEL( dp10dbeta, j, 0 ) = dmu0dJ - 
                                 dp11dmu0*dmu0dJ -
                                 dp11dmu1*dmu1dJ;
	MEL( dp11dbeta, j, 0 ) = dp11dmu0*dmu0dJ +
                                 dp11dmu1*dmu1dJ;
      }
      for( j=0; j<q1; j++ ){
	MEL( dp00dalpha1, j , 0 ) = dp11dpsi01 * psi01 * MEL( z1, 1, j );
	MEL( dp01dalpha1, j , 0 ) = -1.*dp11dpsi01 * psi01 * MEL( z1, 1, j );
	MEL( dp10dalpha1, j , 0 ) = -1.*dp11dpsi01 * psi01 * MEL( z1, 1, j );
	MEL( dp11dalpha1, j , 0 ) = dp11dpsi01 * psi01 * MEL( z1, 1, j );
      }
      for( j=0; j<q2; j++ ){
	MEL( dp00dalpha2, j , 0 ) = 0.0;
	MEL( dp01dalpha2, j , 0 ) = 0.0;
	MEL( dp10dalpha2, j , 0 ) = 0.0;
	MEL( dp11dalpha2, j , 0 ) = 0.0;
      }


      if( n > 2 ){

	for( i=2; i<n; i++ ){
	  
	  y2 = MEL( y, i-2, 0 );
	  mu2 = MEL( mu, i-2, 0 );

	  y1 = MEL( y, i-1, 0 );
	  mu1 = MEL( mu, i-1, 0 );

	  y0 = MEL( y, i, 0 );
	  mu0 = MEL( mu, i, 0 );

	  psi01  = MEL( psi1, i, 0 );
	  gamma1 = log( psi01 );

	  psi02  = MEL( psi2, i, 0 );
	  gamma2 = log( psi02 );

	  /* Recover Delta */

	  delta0 = log( mu0 ) - log( 1.0 - mu0 );

	  converge_delta = 0;
	  count = 0;

	  At = 0.0;

	  do{

	    lpC00 = delta0;
	    lpC01 = delta0 + gamma2;
	    lpC10 = delta0 + gamma1;
	    lpC11 = delta0 + gamma1 + gamma2;

	    mu_current = BoundaryCheck(antilogit( lpC00 ))*p00 +
                         BoundaryCheck(antilogit( lpC01 ))*p01 +
                         BoundaryCheck(antilogit( lpC10 ))*p10 +
                         BoundaryCheck(antilogit( lpC11 ))*p11;

	    At = BoundaryCheck(antilogit( lpC00 )) *
                   BoundaryCheck((1.-antilogit( lpC00 )))*p00 +
                 BoundaryCheck(antilogit( lpC01 )) *
                   BoundaryCheck((1.-antilogit( lpC01 )))*p01 +
                 BoundaryCheck(antilogit( lpC10 )) *
                   BoundaryCheck((1.-antilogit( lpC10 )))*p10 +
                 BoundaryCheck(antilogit( lpC11 )) *
                   BoundaryCheck((1.-antilogit( lpC11 )))*p11;

	    if( fabs( logit(mu_current) - logit(mu0) ) < EPSILON2 ){
	      converge_delta=1;
	    }


	    dstep = (mu0 - mu_current)/( At + 1e-5 );
	    if( dstep > 0.50 ) dstep = 0.50;
	    if( dstep < -0.50 ) dstep = -0.50;

	    delta0 += dstep;

	    if( antilogit(delta0) > 1.0 - NearZero ){
	      delta0 = antilogit( 1.0 - NearZero );
	    }
	    if( antilogit(delta0) < 0.0 + NearZero ){
	      delta0 = antilogit( 0.0 + NearZero );
	    }

	    count++;

	  }while( !converge_delta && count<MAX_COUNT );

	  if( count==MAX_COUNT && !converge_delta ){
	    printf("Failure to recover Delta0\n");
            printf("   mu0 = %f, mu_current = %f \n", mu0, mu_current );
            printf("   At = %f, i = %i \n", At, i );
            printf("   gamma1 = %f, gamma2 = %f\n\n", gamma1, gamma2 );
	    *error_flag += 1;
	  }

	  muTrans = antilogit( delta0 + gamma1*y1 + gamma2*y2 );

	  logLikelihood += log(muTrans)*y0 + log(1.-muTrans)*(1.-y0);

	  MEL( muTc, i, 0 ) = muTrans;

	  for( j=0; j<(p+q1+q2); j++ ){

	    if( j<p ){
	      dDeltaJ = 
                   (   mu0*(1.-mu0)*MEL( x, i, j ) - 
                       (
                       antilogit( lpC00 )*MEL( dp00dbeta, j, 0 ) +
                       antilogit( lpC10 )*MEL( dp10dbeta, j, 0 ) +
                       antilogit( lpC01 )*MEL( dp01dbeta, j, 0 ) +
                       antilogit( lpC11 )*MEL( dp11dbeta, j, 0 ) 
                       ) 
                   )/At;
	      dJ = muTrans*(1.-muTrans)*( dDeltaJ );
	      MEL( dnew00, j, 0 ) = 0.0 -
                    antilogit(lpC00)*(1.-antilogit(lpC00))*dDeltaJ*p00 -
                    antilogit(lpC01)*(1.-antilogit(lpC01))*dDeltaJ*p01 +
		    (1.-antilogit(lpC00))*MEL(dp00dbeta,j,0) +
		    (1.-antilogit(lpC01))*MEL(dp01dbeta,j,0);
	      MEL( dnew01, j, 0 ) = 0.0 -
                    antilogit(lpC10)*(1.-antilogit(lpC10))*dDeltaJ*p10 -
                    antilogit(lpC11)*(1.-antilogit(lpC11))*dDeltaJ*p11 +
		    (1.-antilogit(lpC10))*MEL(dp10dbeta,j,0) +
		    (1.-antilogit(lpC11))*MEL(dp11dbeta,j,0);
	      MEL( dnew10, j, 0 ) = 
                    antilogit(lpC00)*(1.-antilogit(lpC00))*dDeltaJ*p00 +
                    antilogit(lpC01)*(1.-antilogit(lpC01))*dDeltaJ*p01 +
		    antilogit(lpC00)*MEL(dp00dbeta,j,0) +
		    antilogit(lpC01)*MEL(dp01dbeta,j,0);
	      MEL( dnew11, j, 0 ) = 
                    antilogit(lpC10)*(1.-antilogit(lpC10))*dDeltaJ*p10 +
                    antilogit(lpC11)*(1.-antilogit(lpC11))*dDeltaJ*p11 +
		    antilogit(lpC10)*MEL(dp10dbeta,j,0) +
		    antilogit(lpC11)*MEL(dp11dbeta,j,0);
	    }
	    if( j>=p && j<(p+q1) ){
	      dDeltaJ = 0.0 -
                   (
                     (
                     antilogit( lpC00 )*MEL( dp00dalpha1, j-p, 0 ) +
                     antilogit( lpC10 )*MEL( dp10dalpha1, j-p, 0 ) +
                     antilogit( lpC01 )*MEL( dp01dalpha1, j-p, 0 ) +
                     antilogit( lpC11 )*MEL( dp11dalpha1, j-p, 0 ) 
                     ) + (
                     antilogit(lpC10)*(1.-antilogit(lpC10))*
                               MEL(z1,i,j-p)*p10 +
                     antilogit(lpC11)*(1.-antilogit(lpC11))*
                               MEL(z1,i,j-p)*p11
                     )
                   )/At;
	      dJ = muTrans*(1.-muTrans)*( dDeltaJ + y1*MEL( z1, i, j-p ) ); 
	      MEL( dnew00, j, 0 ) = 0.0 -
                    antilogit(lpC00)*(1.-antilogit(lpC00))*(
                       dDeltaJ+0.0*MEL(z1,i,j-p))*p00 -
                    antilogit(lpC01)*(1.-antilogit(lpC01))*(
                       dDeltaJ+0.0*MEL(z1,i,j-p))*p01 +
		    (1.-antilogit(lpC00))*MEL(dp00dalpha1,j-p,0) +
		    (1.-antilogit(lpC01))*MEL(dp01dalpha1,j-p,0);
	      MEL( dnew01, j, 0 ) = 0.0 -
                    antilogit(lpC10)*(1.-antilogit(lpC10))*(
                       dDeltaJ+1.0*MEL(z1,i,j-p))*p10 -
                    antilogit(lpC11)*(1.-antilogit(lpC11))*(
                       dDeltaJ+1.0*MEL(z1,i,j-p))*p11 +
		    (1.-antilogit(lpC10))*MEL(dp10dalpha1,j-p,0) +
		    (1.-antilogit(lpC11))*MEL(dp11dalpha1,j-p,0);
	      MEL( dnew10, j, 0 ) = 
                    antilogit(lpC00)*(1.-antilogit(lpC00))*(
                       dDeltaJ+0.0*MEL(z1,i,j-p))*p00 +
                    antilogit(lpC01)*(1.-antilogit(lpC01))*(
                       dDeltaJ+0.0*MEL(z1,i,j-p))*p01 +
		    antilogit(lpC00)*MEL(dp00dalpha1,j-p,0) +
		    antilogit(lpC01)*MEL(dp01dalpha1,j-p,0);
	      MEL( dnew11, j, 0 ) = 
                    antilogit(lpC10)*(1.-antilogit(lpC10))*(
                       dDeltaJ+1.0*MEL(z1,i,j-p))*p10 +
                    antilogit(lpC11)*(1.-antilogit(lpC11))*(
                       dDeltaJ+1.0*MEL(z1,i,j-p))*p11 +
		    antilogit(lpC10)*MEL(dp10dalpha1,j-p,0) +
		    antilogit(lpC11)*MEL(dp11dalpha1,j-p,0); 
	      }
	    if( j>=(p+q1) ){
	      dDeltaJ = 0.0 -
                   (
                     (
                     antilogit(lpC00)*MEL( dp00dalpha2, j-p-q1, 0 ) +
                     antilogit(lpC10)*MEL( dp10dalpha2, j-p-q1, 0 ) +
                     antilogit(lpC01)*MEL( dp01dalpha2, j-p-q1, 0 ) +
                     antilogit(lpC11)*MEL( dp11dalpha2, j-p-q1, 0 ) 
                     ) + (
                     antilogit(lpC01)*(1.-antilogit(lpC01))* 
                               MEL(z2,i,j-p-q1)*p01 +
                     antilogit(lpC11)*(1.-antilogit(lpC11))*
                               MEL(z2,i,j-p-q1)*p11
                     )
                   )/At;
	      dJ = muTrans*(1.-muTrans)*( dDeltaJ + y2*MEL( z2, i, j-p-q1 ) );
	      MEL( dnew00, j, 0 ) = 0.0 -
                    antilogit(lpC00)*(1.-antilogit(lpC00))*(
                       dDeltaJ+0.0*MEL(z2,i,j-p-q1))*p00 -
                    antilogit(lpC01)*(1.-antilogit(lpC01))*(
                       dDeltaJ+1.0*MEL(z2,i,j-p-q1))*p01 +
		    (1.-antilogit(lpC00))*MEL(dp00dalpha2,j-p-q1,0) +
		    (1.-antilogit(lpC01))*MEL(dp01dalpha2,j-p-q1,0);
	      MEL( dnew01, j, 0 ) = 0.0 -
                    antilogit(lpC10)*(1.-antilogit(lpC10))*(
                       dDeltaJ+0.0*MEL(z2,i,j-p-q1))*p10 -
                    antilogit(lpC11)*(1.-antilogit(lpC11))*(
                       dDeltaJ+1.0*MEL(z2,i,j-p-q1))*p11 +
		    (1.-antilogit(lpC10))*MEL(dp10dalpha2,j-p-q1,0) +
		    (1.-antilogit(lpC11))*MEL(dp11dalpha2,j-p-q1,0);
	      MEL( dnew10, j, 0 ) = 
                    antilogit(lpC00)*(1.-antilogit(lpC00))*(
                       dDeltaJ+0.0*MEL(z2,i,j-p-q1))*p00 +
                    antilogit(lpC01)*(1.-antilogit(lpC01))*(
                       dDeltaJ+1.0*MEL(z2,i,j-p-q1))*p01 +
		    antilogit(lpC00)*MEL(dp00dalpha2,j-p-q1,0) +
		    antilogit(lpC01)*MEL(dp01dalpha2,j-p-q1,0);
	      MEL( dnew11, j, 0 ) = 
                    antilogit(lpC10)*(1.-antilogit(lpC10))*(
                       dDeltaJ+0.0*MEL(z2,i,j-p-q1))*p10 +
                    antilogit(lpC11)*(1.-antilogit(lpC11))*(
                       dDeltaJ+1.0*MEL(z2,i,j-p-q1))*p11 +
		    antilogit(lpC10)*MEL(dp10dalpha2,j-p-q1,0) +
		    antilogit(lpC11)*MEL(dp11dalpha2,j-p-q1,0);
	    }

	    MEL( score, j, 0 ) += dJ*( y0 - muTrans )/(muTrans*(1.-muTrans));

	    for( k=0; k<(p+q1+q2); k++ ){

	      if( k<p ){
		dDeltaK = 
                   (   mu0*(1.-mu0)*MEL( x, i, k ) - 
                       (
                       antilogit( lpC00 )*MEL( dp00dbeta, k, 0 ) +
                       antilogit( lpC10 )*MEL( dp10dbeta, k, 0 ) +
                       antilogit( lpC01 )*MEL( dp01dbeta, k, 0 ) +
                       antilogit( lpC11 )*MEL( dp11dbeta, k, 0 ) 
                       ) 
                   )/At;
		dK = muTrans*(1.-muTrans)*( dDeltaK );
	      }
	      if( k>=p && k<(p+q1) ){
		dDeltaK = 0.0 -
                   (
                     (
                     antilogit( lpC00 )*MEL( dp00dalpha1, k-p, 0 ) +
                     antilogit( lpC10 )*MEL( dp10dalpha1, k-p, 0 ) +
                     antilogit( lpC01 )*MEL( dp01dalpha1, k-p, 0 ) +
                     antilogit( lpC11 )*MEL( dp11dalpha1, k-p, 0 ) 
                     ) + (
                     antilogit(lpC10)*(1.-antilogit(lpC10))*
                               MEL(z1,i,k-p)*p10 +
                     antilogit(lpC11)*(1.-antilogit(lpC11))*
                               MEL(z1,i,k-p)*p11
                     )
                   )/At;
		dK = muTrans*(1.-muTrans)*( dDeltaK + y1*MEL( z1, i, k-p ) ); 
	      }
	      if( k>=(p+q1) ){
		dDeltaK = 0.0 -
                   (
                     (
                     antilogit(lpC00)*MEL( dp00dalpha2, k-p-q1, 0 ) +
                     antilogit(lpC10)*MEL( dp10dalpha2, k-p-q1, 0 ) +
                     antilogit(lpC01)*MEL( dp01dalpha2, k-p-q1, 0 ) +
                     antilogit(lpC11)*MEL( dp11dalpha2, k-p-q1, 0 ) 
                     ) + (
                     antilogit(lpC01)*(1.-antilogit(lpC01))*
                               MEL(z2,i,k-p-q1)*p01 +
                     antilogit(lpC11)*(1.-antilogit(lpC11))*
                               MEL(z2,i,k-p-q1)*p11
                     )
                   )/At;
		dK = muTrans*(1.-muTrans)*(dDeltaK + y2*MEL( z2, i, k-p-q1 ));
	      }

	      MEL( hessian, j, k ) += dJ * dK / ( muTrans*(1.-muTrans) );

	    }/* k loop */

	  }/* j loop */

	  /* update the probabilities and derivatives */

	  new00 = (1.-antilogit(lpC00))*p00 +
                  (1.-antilogit(lpC01))*p01;
	  new10 = antilogit(lpC00)*p00 +
                  antilogit(lpC01)*p01;
	  new01 = (1.-antilogit(lpC10))*p10 +
                  (1.-antilogit(lpC11))*p11;
	  new11 = antilogit(lpC10)*p10 +
                  antilogit(lpC11)*p11;

	  p00 = new00;
	  p01 = new01;
	  p10 = new10;
	  p11 = new11;

	  for( j=0; j<p; j++ ){
	    MEL( dp00dbeta, j, 0 ) = MEL( dnew00, j, 0 );
	    MEL( dp01dbeta, j, 0 ) = MEL( dnew01, j, 0 );
	    MEL( dp10dbeta, j, 0 ) = MEL( dnew10, j, 0 );
	    MEL( dp11dbeta, j, 0 ) = MEL( dnew11, j, 0 );
	  }
	  for( j=0; j<q1; j++ ){
	    MEL( dp00dalpha1, j, 0 ) = MEL( dnew00, p+j, 0 );
	    MEL( dp01dalpha1, j, 0 ) = MEL( dnew01, p+j, 0 );
	    MEL( dp10dalpha1, j, 0 ) = MEL( dnew10, p+j, 0 );
	    MEL( dp11dalpha1, j, 0 ) = MEL( dnew11, p+j, 0 );
	  }
	  for( j=0; j<q2; j++ ){
	    MEL( dp00dalpha2, j, 0 ) = MEL( dnew00, p+q1+j, 0 );
	    MEL( dp01dalpha2, j, 0 ) = MEL( dnew01, p+q1+j, 0 );
	    MEL( dp10dalpha2, j, 0 ) = MEL( dnew10, p+q1+j, 0 );
	    MEL( dp11dalpha2, j, 0 ) = MEL( dnew11, p+q1+j, 0 );
	  }

	  /* score test for lag-3 */
	  if( n>3 ){
		if(i > 2) {  /* only count lag-3 measurements */
			y3 = MEL( y, i-3, 0 );
			*U2c += y3*(y0-muTrans);
		}
	  }

	}/* i loop */

      }/* n>2 */

  } /* end n>1 */

  destroy_matrix( dp00dbeta );
  destroy_matrix( dp01dbeta );
  destroy_matrix( dp10dbeta );
  destroy_matrix( dp11dbeta );

  destroy_matrix( dp00dalpha1 );
  destroy_matrix( dp01dalpha1 );
  destroy_matrix( dp10dalpha1 );
  destroy_matrix( dp11dalpha1 );

  destroy_matrix( dp00dalpha2 );
  destroy_matrix( dp01dalpha2 );
  destroy_matrix( dp10dalpha2 );
  destroy_matrix( dp11dalpha2 );

  destroy_matrix( dnew00 );
  destroy_matrix( dnew01 );
  destroy_matrix( dnew10 );
  destroy_matrix( dnew11 );

  M = (MATRIX **)calloc( 2, (unsigned)sizeof( struct matrix ) );

  /* wfile=fopen("MLE.out","a");
fprintf(wfile," score\n");
fmatdump(wfile,score);
fprintf(wfile," hessian\n");
fmatdump(wfile,hessian);
fprintf(wfile," \n");
fclose(wfile);  */

  M[0] = score;
  M[1] = hessian;

  *logLc = logLikelihood;

  return M;

}
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */

void mtm_lag1( S_x, S_y, S_z, S_id, S_beta, S_alpha, S_logL, 
	     S_nobs, S_nclust, S_p, S_q, S_tol, S_covM, S_covE,
             S_test, S_muT, S_offset )

  double  *S_x, *S_y, *S_z, *S_id, *S_beta, *S_alpha, *S_covM, *S_covE;
  double  *S_offset;
  double  *S_test, *S_muT, *S_logL, *S_tol;
  integer *S_nobs, *S_nclust, *S_p, *S_q;
{
  MATRIX  **Y, **X, **Z, **MuTrans, **Offset;
  MATRIX  *y, *mu, *beta, *alpha, *z, *x, *id, *muT, *off;
  double  dmax, tolerance, logL, logLc, ccc, ScoreTest, U2, V2, U2c, V2c;
  MATRIX  *delta, *DVD, *DVR;
  MATRIX  **M ;
  MATRIX  *UUt, *Ie, *Im;
  int     i, j, k, n, nobs, nclust, p, q, one, nc2, count, iter, c;
  int     method, converge;
  FILE    *wfile;

  one = 1;

  p = *S_p;
  q = *S_q;
  nobs = *S_nobs;
  nclust = *S_nclust;
  tolerance = *S_tol;

  logL = *S_logL;

  wfile = fopen("MLE.out","w");
  fprintf( wfile, "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
  fprintf( wfile, " p = %i, q = %i, n = %i, tol = %f\n", p, q, nobs, 
	   tolerance );
  fprintf( wfile, "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n\n" );
  fclose( wfile );


  from_S( S_beta, S_p, &one, beta );
  make_permanent( beta );

  from_S( S_alpha, S_q, &one, alpha );
  make_permanent( alpha );

  from_S( S_x, S_nobs, S_p, x );
  from_S( S_y, S_nobs, &one, y );
  from_S( S_z, S_nobs, S_q, z );
  from_S( S_muT, S_nobs, &one, muT);
  from_S( S_id, S_nobs, &one, id );
  from_S( S_offset, S_nobs, &one, off );

  X = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  Y = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  Z = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  MuTrans = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );
  Offset = (MATRIX **)calloc( nclust, (unsigned)sizeof( struct matrix) );

  make_permanent( id );

  split( x, id, X );
  split( y, id, Y );
  split( z, id, Z );
  split( muT, id, MuTrans );
  split( off, id, Offset );

  destroy_matrix( x );
  destroy_matrix( y );
  destroy_matrix( z );
  destroy_matrix( id );
  destroy_matrix( muT );
  destroy_matrix( off );

  for(c=0;c<nclust;c++){
    make_permanent( Y[c] );
    make_permanent( X[c] );
    make_permanent( Z[c] ); 
    make_permanent( Offset[c] ); }


  iter=0;
  converge=0;


  do{

    DVR = create_matrix( p+q, 1, EPHEMERAL );
    DVD = create_matrix( p+q, p+q, EPHEMERAL );
    UUt = create_matrix( p+q, p+q, EPHEMERAL );

    logL = 0.0;
    U2 = 0.0;
    V2 = 0.0;

    for(c=0;c<nclust;c++){

      n = Y[c]->nrows;

      logLc = 0.0;
      U2c = 0.0;
      V2c = 0.0;

      M = get_score_calcs1( Y[c], X[c], Z[c], beta, alpha, n , 
                           &logLc, &U2c, &V2c, MuTrans[c], Offset[c]  );

      make_permanent( M[0] );
      UUt = matadd( UUt, matmult( M[0], transp( M[0] ) ) );
      make_ephemeral( M[0] );
      DVR = matadd( DVR, M[0] );
      DVD = matadd( DVD, M[1] );

      cfree( M );
  
      logL += logLc;
      U2 += U2c;
      /*      V2 += V2c; */
      V2 += U2c*U2c;

    }/* c */

    make_permanent( DVD );
    make_permanent( DVR );

    delta = matmult( luinv( DVD ), DVR );

wfile=fopen("MLE.out","a");
fprintf(wfile," score\n");
fmatdump(wfile, DVR );
fprintf(wfile," hessian\n");
fmatdump(wfile,DVD);
fprintf(wfile," delta\n");
fmatdump(wfile,delta);
fprintf(wfile," logL = %f\n", logL );
fclose(wfile); 

    make_permanent( delta );
    dmax = matmaxabs( delta );

    ccc = 1.0;
    if( iter<10 ) ccc = 0.5;

    for( j=0; j<p; j++) MEL( beta, j, 0 ) += ccc*MEL( delta, j, 0 );
    for( j=0; j<q; j++) MEL( alpha, j, 0 ) += ccc*MEL( delta, p+j, 0 );
    destroy_matrix( delta );

    if( dmax < tolerance ){
      converge = 1;
    }else{
      destroy_matrix( DVD );
      destroy_matrix( DVR );
      destroy_matrix( UUt );
    }

    iter++;

 }while( (iter < MAXITS) && (!converge) );                     


  /* add the appropriate cleaning... */
  count = 0;
  for( i=0; i<nclust; i++ ){
    for( j=0; j<MuTrans[i]->nrows; j++ ){
      *(S_muT+count) = MEL( MuTrans[i], j, 0 );
      count++;
    }
    destroy_matrix( MuTrans[i] );
    destroy_matrix( Y[i] );
    destroy_matrix( X[i] );
    destroy_matrix( Z[i] );
  }
  cfree( MuTrans );
  cfree( Y );
  cfree( X );
  cfree( Z );

  to_S( beta, S_beta );

  to_S( alpha, S_alpha );

  Im = luinv( DVD );
  make_permanent( Im );

  Ie = matmult( Im, matmult( UUt, transp(Im) ) );

  to_S( Im, S_covM );

  to_S( Ie, S_covE );

  *S_logL = logL;

  ScoreTest = U2*U2/V2;

  *S_test = ScoreTest;

}/* end of routine */

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */

MATRIX   **get_score_calcs1( y, x, z, beta, alpha, n, logLc, U2c, V2c, muTc,
                            offset )

  MATRIX *y, *x, *z, *beta, *alpha, *muTc, *offset;
  double *logLc, *U2c, *V2c;
  int    n;

{
  MATRIX *score, *hessian;
  MATRIX **M, *mu, *psi, *lp1, *lp2;
  double mu2, mu1, psi21, p21, y2, y1;
  double A, B, C, R, muTrans;
  double dp21dmu1, dp21dmu2, dp21dpsi21;
  double dlogLcdmu1, dlogLcdmu2, dlogLcdpsi21;
  double dmuTransdmu1, dmuTransdmu2, dmuTransdpsi21;
  double dmu1dthetaJ, dmu2dthetaJ, dpsi21dthetaJ;
  double dmu1dthetaK, dmu2dthetaK, dpsi21dthetaK;
  double dJ, dK;
  double mu0, pi11, pi01, muT0, muT1, varEst;
  double logLikelihood;
  int    p, q, i, j, k;
  FILE   *wfile;

  p = beta->nrows;
  q = alpha->nrows;

  /* create the score and hessian matrices */

  score = create_matrix( p+q, 1, EPHEMERAL );
  hessian = create_matrix( p+q, p+q, EPHEMERAL );

  logLikelihood = 0.0;

  /* deal with scalar quantities */
  
  mu = create_matrix( n, 1, EPHEMERAL );
  psi = create_matrix( n, 1, EPHEMERAL );

  lp1 = matmult( x, beta );
  lp1 = matadd( lp1, offset );
  lp2 = matmult( z, alpha );
  
  for(i=0;i<n;i++){
    MEL( mu, i, 0 ) = antilogit( MEL(lp1,i,0) );
    MEL( psi, i, 0 ) = exp( MEL(lp2,i,0) );
  }

  destroy_matrix( lp1 );
  destroy_matrix( lp2 );


  /*  first measurement */

  y2 = MEL( y, 0, 0 );
  mu2 = MEL( mu, 0, 0 );
  
  for( j=0; j<p; j++ ){
     MEL( score, j, 0 ) += MEL( x, 0, j )*(y2-mu2);
     for( k=0; k<p; k++ ){
       MEL( hessian, j, k ) += MEL( x, 0, j )*mu2*(1-mu2)*MEL( x, 0, k);
     }
  }

  logLikelihood += y2*log(mu2) + (1.-y2)*log(1.-mu2); 

  MEL( muTc, 0, 0 ) = mu2;

  /* each transition */

  if( n>1 ){

    for( i=1; i<n; i++ ){
 
      y2 = MEL( y, i, 0 );
      y1 = MEL( y, i-1, 0 );

      mu2 = MEL( mu, i, 0 );
      mu1 = MEL( mu, i-1, 0 );
      psi21 = MEL( psi, i, 0 );

      if( fabs( psi21 - 1.0 ) < EPSILON1 ){

	p21 = mu2*mu1;

	dp21dmu1 = mu2;

	dp21dmu2 = mu1;

	dp21dpsi21 = 0.0;

      }else{

        A = (psi21-1.);
	B = 1. + (mu1+mu2)*(psi21-1.);
	C = psi21*mu1*mu2;
        R = sqrt( B*B - 4.*A*C );

        p21 = ( B - R )/( 2.*A );
	
	dp21dmu1 = ( (psi21-1.) - (0.5/R)*( 2.*B*(psi21-1.) -
                      4.*A*psi21*mu2 ) )/( 2.*A ); 

	dp21dmu2 = ( (psi21-1.) - (0.5/R)*( 2.*B*(psi21-1.) -
                      4.*A*psi21*mu1 ) )/( 2.*A ); 

        dp21dpsi21 = ( B - R )*( -1./( 2*A*A ) ) +
                     ( ( mu1+mu2 - 
                       (0.5/R)*( 2.*B*(mu1+mu2) -
		        4.*mu1*mu2*( 2.*psi21 - 1.) ) ) )/( 2.*A ); 
      }

      muTrans = ( p21 / mu1 )*y1 + ( (mu2-p21)/(1.-mu1) )*(1.-y1);

      logLikelihood += log(muTrans)*y2 + log(1.-muTrans)*(1.-y2);

      MEL( muTc, i, 0 ) = muTrans;

      dmuTransdmu1 =( p21 * y1 )*( -1./( mu1*mu1 ) ) +
                    ( y1 / mu1 )*dp21dmu1 +
                    ( (mu2-p21)*(1.-y1) )*( 1./( (1.-mu1)*(1.-mu1) ) ) +
                    ( (1.-y1)/(1.-mu1) )*(-1.*dp21dmu1);

      dlogLcdmu1 = dmuTransdmu1 * ( y2 - muTrans )/( muTrans*(1.-muTrans) );


      dmuTransdmu2 = ( y1/mu1 )*dp21dmu2 + 
                     ( (1.-y1)/(1.-mu1) ) * ( 1. - dp21dmu2 );

      dlogLcdmu2 = dmuTransdmu2 * ( y2 - muTrans )/( muTrans*(1.-muTrans) );


      dmuTransdpsi21 = ( y1 / mu1 )*dp21dpsi21 - 
                       ( (1.-y1)/(1.-mu1) )*dp21dpsi21; 

      dlogLcdpsi21 = dmuTransdpsi21*( y2 - muTrans )/( muTrans*(1.-muTrans) );

      for( j=0; j<(p+q); j++ ){
	
	if( j < p ){
	  dmu1dthetaJ = mu1*(1.-mu1)*MEL(x,i-1,j);
	  dmu2dthetaJ= mu2*(1.-mu2)*MEL(x,i,j);
	  dpsi21dthetaJ = 0.0;
	}else{
	  dmu1dthetaJ = 0.0; 
	  dmu2dthetaJ= 0.0;
	  dpsi21dthetaJ = psi21 * MEL(z,i,j-p);
	}

        dJ = dmuTransdmu1 * dmu1dthetaJ +
             dmuTransdmu2 * dmu2dthetaJ +
             dmuTransdpsi21 * dpsi21dthetaJ;

	MEL( score, j, 0 ) += dlogLcdmu1 * dmu1dthetaJ +
                              dlogLcdmu2 * dmu2dthetaJ +
                              dlogLcdpsi21 * dpsi21dthetaJ;

	for( k=0; k<(p+q); k++ ){

	  if( k < p ){
	    dmu1dthetaK = mu1*(1.-mu1)*MEL(x,i-1,k);
	    dmu2dthetaK = mu2*(1.-mu2)*MEL(x,i,k);
	    dpsi21dthetaK = 0.0;
	  }else{
	    dmu1dthetaK = 0.0; 
	    dmu2dthetaK = 0.0;
	    dpsi21dthetaK = psi21 * MEL(z,i,k-p);
	  }

	  dK = dmuTransdmu1 * dmu1dthetaK +
	       dmuTransdmu2 * dmu2dthetaK +
               dmuTransdpsi21 * dpsi21dthetaK;

	  MEL( hessian, j, k ) += dJ * dK / ( muTrans*(1.-muTrans) );
	}
      }

      /* score test calculations */

      if( n>2 ){

	*U2c += MEL(y,i-2,0)*( MEL(y,i,0) - muTrans );

	mu1 = MEL( mu, i-1, 0 );
	mu0 = MEL( mu, i-2, 0 );

	psi21 = MEL( psi, i-1, 0 );

	if( fabs( psi21 - 1.0 ) < EPSILON1 ){

	  pi11 = mu0*mu1;

	}else{

	  A = (psi21-1.);
	  B = 1. + (mu1+mu0)*(psi21-1.);
	  C = psi21*mu1*mu0;
	  R = sqrt( B*B - 4.*A*C );

	  pi11 = ( B - R )/( 2.*A );

	}

	pi01 = (mu0-pi11)/mu0;
	pi11 = pi11/mu0;

	muT1 = p21/mu1;
	muT0 = (mu2-p21)/(1.-mu1);

	varEst = muT1*(1.-muT1)*pi11 + muT0*(1.-muT0)*pi01;

        *V2c += varEst;

      }/* end n>2 */
      

    }/* end i loop */

  }/* end n>1 */


  M = (MATRIX **)calloc( 2, (unsigned)sizeof( struct matrix ) );

  /* wfile=fopen("MLE.out","a");
fprintf(wfile," score\n");
fmatdump(wfile,score);
fprintf(wfile," hessian\n");
fmatdump(wfile,hessian);
fprintf(wfile," \n");
fclose(wfile);  */

  M[0] = score;
  M[1] = hessian;

  *logLc = logLikelihood;

  return M;

}

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */
double antilogit( x )

  double x;
{
  double out;

  out = exp( x )/(1.0+exp( x ) );

  return out;

}
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

double logit( x )

  double x;
{
  double out;

  if( ( x < NearZero) | (1.0-x < NearZero) ){
    out=0.0;
  }else{
    out = log( x ) - log( 1.0 - x );
  }

  return out;

}
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

double BoundaryCheck( x )

  double x;
{
  double out;

  out = x;

  if( out < NearZero ) out = NearZero;
  if( out > 1.0 - NearZero ) out = 1.0 - NearZero;

  return out;

}

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */







